// Arduino sketch for Penny, the 3 servo hexapod By: JerZ
// Visit www.LetsMakeRobots.com/node/38472 for more info.

// Robot designed for CheapDuino(Arduino compatible) microcontroller
// Special thanks to DFRobot.com for making this robot possible.


// LINK REQUIRED LIBRARY //
#include <Servo.h>        // Link to servo library


// PIN SETUP //
const byte IRLEDPin     =  A0;   // IR LEDs on single pin
const byte lSensorPin   =  A4;   // IR sensor for left eye
const byte rSensorPin   =  A5;   // IR sensor for right eye
const byte lServoPin    =   9;   // Servo pin for left legs
const byte rServoPin    =  10;   // Servo pin for right legs
const byte cServoPin    =  11;   // Servo pin for center legs
const byte whiteLED     =  13;   // Pin for onboard white LED


// CREATE INSTANCES OF SERVO OBJECTS //
Servo leftLegs;
Servo rightLegs;
Servo centerLegs;


// IR SENSOR VARIABLES //
unsigned long PulseOffTimer;     // Timer for IR LED sensor (off time)
unsigned long PulseOnTimer;      // Timer for IR LED sensor (on time)
const int IRPulse        =  25;  // Duration IR LED pulses
const int IROff          = 225;  // Time between IR LED pulses
const int lTrigger       = 400;  // Trigger value for left IR sensor (Range is about 100-500 for the schematic posted on LMR.)
const int rTrigger       = 400;  // Trigger value for right IR sensor (Higher values will increase sensor range)
int lSensorRaw, rSensorRaw;      // Left/Right IR sensor data  (Typically 70-650 your results may vary)


// SERVO VARIABLES //
byte lServoAngle         =  90;  // Angle of left leg servo (
const byte lServoMIN     =  30;  // Mechanical limits of left leg servo
const byte lServoMAX     = 120;  // Left leg servo range (40-130, 90 degrees)

byte rServoAngle         =  90;  // Angle of right leg servo
const byte rServoMIN     =  30;  // Mechanical limits of right leg servo
const byte rServoMAX     = 120;  // Right leg servo range (30-120, 90 degrees)

byte cServoAngle         =  90;  // Angle of center leg servo
const byte cServoMIN     =  50;  // Mechanical limits of center leg servo
const byte cServoMAX     =  80;  // Center leg servo range (60-80, 20 degrees)


// STATE MACHINE VARIABLES //
const byte forwardState  =   0;
const byte reverseState  =   1;
const byte leftState     =   2;
const byte rightState    =   3;

boolean stepPhase        =   0;  // Flag indicating which half of step cycle we are in
const byte servoDelay    =  10;  // Set delay (in milliseconds) between servo positions
const byte servoSpeed    =   3;  // Distance in degrees that servos will move in each program loop
byte state, oldState     =   0;  // Stores current/previous mode value

const byte dwellDuration =   6;  // Number of half steps robot will stay in turn/back-up mode
byte dwellTimer          =   0;  // Keeps track of number of "ticks" remaining in a "dwell" state
byte dwellTotalTicks     =   ((lServoMAX - lServoMIN) / servoSpeed) * dwellDuration;  // Number of "frames" in a complete step * number of steps to dwell


// PROGRAM SETUP //
void setup(){
  pinMode(IRLEDPin, OUTPUT);     // Initialize pins (Servo pins are set as "OUTPUT" inside Servo Library with "attach")
  pinMode(lSensorPin, INPUT);
  pinMode(rSensorPin, INPUT);
  leftLegs.attach(lServoPin);    // Attach servos to pins
  rightLegs.attach(rServoPin);
  centerLegs.attach(cServoPin);
}


// PROGRAM MAIN LOOP //
void loop(){
  pulseController();           // Function that drives the IR LED sensor pulses
  if(IRLEDPin) readSensors();  // Poll IR sensors for raw data if IR LEDs are on
  stateMachine();
  switch(state){               // Determine which mode we are currently in
    case forwardState:
      moveForward();           // Call the function to walk forward
    break;
    case reverseState:
      moveReverse();           // Call the function to walk backward
    break;
    case leftState:
      moveTurnLeft();          // Call the function to turn left
    break;
    case rightState:
      moveTurnRight();         // Call the function to turn right
    break;
  }  
}


// FUNCTIONS //
void pulseController(){
  unsigned long currMillis = millis();
  if(currMillis - PulseOffTimer > IROff){                                 // Check to see if it's time to switch IR LEDs on
    digitalWrite(IRLEDPin,HIGH);                                          // Turn on IR LEDs
    PulseOnTimer = currMillis;                                            // Record time IR LEDs got switched on
    delay(5);                                                             // Give IR LEDs some time to stabilize
  }
  if(digitalRead(IRLEDPin)==HIGH && currMillis - PulseOnTimer > IRPulse){ // Check to see if it's time to turn IR LEDs off
    digitalWrite(IRLEDPin,LOW);                                           // Turn off IR LEDs
    PulseOffTimer = currMillis;                                           // Record time IR LEDs got switched off
  }
}


void readSensors(){
  lSensorRaw = analogRead(lSensorPin);                      // Read and store left sensor data
  rSensorRaw = analogRead(rSensorPin);                      // Read and store right sensor data
}


void stateMachine(){
  unsigned long currMillis = millis();
  
  if(lSensorRaw > lTrigger && rSensorRaw > rTrigger){          // No obstacles are detected

    state = forwardState;                                      // Set mode to walk forward
  }else if(lSensorRaw <= lTrigger && rSensorRaw <= rTrigger){  // Set default both sensors triggered incase there is an error

    state = reverseState;                                      // Set mode to reverse
  }else if(lSensorRaw > lTrigger && rSensorRaw <= rTrigger){   // Right sensor detects obstacle

    state = leftState;                                         // Set mode to turn left
  }else if(lSensorRaw <= lTrigger && rSensorRaw > rTrigger){   // Left sensor detects obstacle

    state = rightState;                                        // Set mode to turn right
  }
}


void moveForward(){                        // Servo routine for walking forward (Exactly the same as "moveBackward" except center leg position is inverted)

  if (!stepPhase){                         // First half of step cycle (Left leg extend, right leg retract)
    cServoAngle = cServoMIN;               // Set position of center legs for this half of the cycle
    if (lServoAngle < lServoMAX){          // Make sure legs aren't at their limit
      lServoAngle += servoSpeed;           // Move legs "servoSpeed" degrees once per loop
      rServoAngle += servoSpeed;
    }else {
      lServoAngle = lServoMAX;             // If a leg is at its limit, toggle stepPhase to run second half of step cycle
      rServoAngle = rServoMAX;
      stepPhase = !stepPhase;
    }
    centerLegs.write(cServoAngle);         // Write new positions to servos
    leftLegs.write(lServoAngle);
    rightLegs.write(rServoAngle);
  }else {                                  // Second half of step cycle (Left leg retract, right leg extend)
    cServoAngle = cServoMAX;               // Set position of center legs for this half of the cycle
    if (lServoAngle > lServoMIN){
      lServoAngle -= servoSpeed;           // Move legs "servoSpeed" degrees once per loop
      rServoAngle -= servoSpeed;
    }else {
      lServoAngle = lServoMIN;             // If a leg is at its limit, toggle stepPhase to run first half of step cycle
      rServoAngle = rServoMIN;
      stepPhase = !stepPhase;  
    }
    centerLegs.write(cServoAngle);
    leftLegs.write(lServoAngle);           // Write new positions to servos
    rightLegs.write(rServoAngle);
  }
  delay(servoDelay);                       // Wait so movement is slowed and smooth
}


void moveReverse(){                        // Servo routine for walking backward (Exactly the same as "moveForward" except center leg position is inverted)
  dwellTimer = dwellTotalTicks;            // Set a timer to help the robot from getting stuck at the edge of its' sensor range
  while(dwellTimer){                       // Lock state inside loop so it can't be interupted until "dwellTimer" runs out
    if (stepPhase){                        // First half of step cycle (Left leg extend, right leg retract)
      cServoAngle = cServoMAX;             // Set position of center legs for this cycle
      if (lServoAngle < lServoMAX){
        lServoAngle += servoSpeed;         // Move legs "servoSpeed" degrees once per loop
        rServoAngle += servoSpeed;
      }else {
        lServoAngle = lServoMAX;           // If a leg is at its limit, toggle stepPhase to run second half of step cycle
        rServoAngle = rServoMAX;
        stepPhase = !stepPhase;
      }
      centerLegs.write(cServoAngle);       // Write new positions to servos
      leftLegs.write(lServoAngle);
      rightLegs.write(rServoAngle);
    }else {                                // Second half of step cycle (Left leg retract, right leg extend)
      cServoAngle = cServoMIN;             // Set position of center legs for this cycle
      if (lServoAngle > lServoMIN){
        lServoAngle -= servoSpeed;         // Move leg "servoSpeed" degrees once per loop
        rServoAngle -= servoSpeed;
      }else {
        lServoAngle = lServoMIN;           // If a leg is at its limit, toggle stepPhase to run first half of step cycle
        cServoAngle = cServoMIN;
        stepPhase = !stepPhase;
      }
      centerLegs.write(cServoAngle);       // Write new positions to servos
      leftLegs.write(lServoAngle);
      rightLegs.write(rServoAngle);
    }
    delay(servoDelay);                     // Wait so movement is slowed and smooth
    dwellTimer--;                          // Subtract from "dwellTimer" and start "while()" loop again
  }
}


void moveTurnLeft(){                       // Servo routine for turning left (Exactly the same as moveTurnRight except center leg position is inverted)
  dwellTimer = dwellTotalTicks;            // Set a timer to help the robot from getting stuck at the edge of its' sensor range
  while(dwellTimer){                       // Lock state inside loop so it can't be interupted until "dwellTimer" runs out
    if (stepPhase){                        // First half of step cycle (Left leg extend, right leg extend)
      cServoAngle = cServoMAX;             // Set position of center legs for this cycle
      if (lServoAngle < lServoMAX){
        lServoAngle += servoSpeed;         // Move legs "servoSpeed" degrees once per loop
        rServoAngle -= servoSpeed;
      }else {
        lServoAngle = lServoMAX;           // If a leg is at its limit, toggle stepPhase to run second half of step cycle
        rServoAngle = rServoMIN;
        stepPhase = !stepPhase;
      }
      centerLegs.write(cServoAngle);
      leftLegs.write(lServoAngle);         // Write new positions to servos
      rightLegs.write(rServoAngle);
    }else {                                // Second half of step cycle (Left leg retract, right leg retract)
      cServoAngle = cServoMIN;             // Set position of center legs for this cycle
      if (lServoAngle > lServoMIN){
        lServoAngle -= servoSpeed;         // Move leg "servoSpeed" degrees once each program loop
        rServoAngle += servoSpeed;
      }else {
        lServoAngle = lServoMIN;           // If a leg is at its limit, toggle stepPhase to run first half of step cycle
        rServoAngle = rServoMAX;
        stepPhase = !stepPhase;  
      }
      centerLegs.write(cServoAngle);       // Write new positions to servos
      leftLegs.write(lServoAngle);
      rightLegs.write(rServoAngle);
    }
    delay(servoDelay);                     // Wait so movement is slowed and smooth
    dwellTimer--;                          // Subtract from "dwellTimer" and start "while()" loop again
  }
}


void moveTurnRight(){                      // Servo routine for turning right (Exactly the same as moveTurnLeft except center leg position is inverted)
  dwellTimer = dwellTotalTicks;            // Set a timer to help the robot from getting stuck at the edge of its' sensor range
  while(dwellTimer){                       // Lock state inside loop so it can't be interupted until "dwellTimer" runs out
    if (!stepPhase){                       // First half of step cycle (Left leg extend, right leg extend)
      cServoAngle = cServoMIN;             // Set position of center legs for this cycle
      if (lServoAngle < lServoMAX){
        lServoAngle += servoSpeed;         // Move legs "servoSpeed" degrees once per loop
        rServoAngle -= servoSpeed;
      }else {
        lServoAngle = lServoMAX;           // If a leg is at its limit, toggle stepPhase to run second half of step cycle
        rServoAngle = rServoMIN;
        stepPhase = !stepPhase;
      }
      centerLegs.write(cServoAngle);
      leftLegs.write(lServoAngle);         // Write new positions to servos
      rightLegs.write(rServoAngle);
    }else {                                // Second half of step cycle (Left leg retract, right leg retract)
      cServoAngle = cServoMAX;             // Set position of center legs for this cycle
      if (lServoAngle > lServoMIN){
        lServoAngle -= servoSpeed;         // Move leg "servoSpeed" degrees once each program loop
        rServoAngle += servoSpeed;
      }else {
        lServoAngle = lServoMIN;           // If a leg is at its limit, toggle stepPhase to run first half of step cycle
        rServoAngle = rServoMAX;
        stepPhase = !stepPhase;   
      }
      centerLegs.write(cServoAngle);       // Write new positions to servos
      leftLegs.write(lServoAngle);
      rightLegs.write(rServoAngle);
    }
    delay(servoDelay);                     // Wait so movement is slowed and smooth
    dwellTimer--;                          // Subtract from "dwellTimer" and start "while()" loop again
  }
}
